# -*- coding: utf-8 -*-

import textwrap
import uritemplate

from django.utils.encoding import force_text
from django.utils.functional import cached_property

from rest_framework import serializers, fields, viewsets, relations, status

from http.client import responses


def docstring_beautify(docstring):
    if docstring is None:
        return ''
    return textwrap.fill(docstring.strip())


class BaseParameter(object):
    format = None
    is_complex = False

    def __init__(self, name, field, mode=None):
        self.name = name
        self.field = field

    def __eq__(self, other):
        return (isinstance(other, BaseParameter) and
                self.name == other.name and
                type(self.field) is type(other.field))  # noqa

    @property
    def description(self):
        return force_text(self.field.help_text or '')

    def __repr__(self):
        return '{}<{}>({})'.format(self.__class__.__name__, self.name, self.field)

    @property
    def type(self):
        raise NotImplementedError(self.__class__, repr(self))

    def get_schema(self):
        value = {
            'description': self.description,
            'type': self.type,
            'allowEmptyValue': self.field.allow_null,
        }
        if self.name:
            value['name'] = self.name
        if isinstance(self.field, serializers.ChoiceField):
            value['enum'] = list(self.field.choices.keys())
        return value


class ManyParameter(BaseParameter):
    def __init__(self, name, field, mode='write'):
        super().__init__(name, field, mode='write')
        if isinstance(field, relations.ManyRelatedField):
            child = field.child_relation
        else:
            child = field.child
        cls = get_parameter_class(child)
        self.items = cls(None, child, mode=mode)

    @property
    def type(self):
        return 'array'

    @property
    def is_complex(self):
        return isinstance(self.field, (serializers.Serializer,
                                       serializers.ListSerializer))

    def get_schema(self):
        s = super().get_schema()
        s['items'] = self.items.get_schema()
        return s


class Parameter(BaseParameter):
    type_mapping = {
        fields.IntegerField: 'integer',
        fields.FloatField: 'number',
        fields.BooleanField: 'boolean',
        fields.CharField: 'string',
        fields.DateTimeField: 'string',
        fields.DateField: 'string',
    }
    format_mapping = {
        fields.DateTimeField: 'dateTime',
        fields.DateField: 'date',
    }

    def get_schema(self):
        s = super().get_schema()
        if type(self.field) in self.format_mapping:
            s['format'] = self.format_mapping[type(self.field)]
        return s

    @property
    def type(self):
        return self.type_mapping.get(type(self.field), 'string')


class SubParameter(BaseParameter):
    type = 'object'

    def __init__(self, name, field, mode='write'):
        super().__init__(name, field)
        self.body = Body.from_serializer(field, mode=mode)

    def get_schema(self):
        s = super().get_schema()
        s.update(self.body.get_schema())
        return s


def get_parameter_class(field):
    if isinstance(field, (serializers.ListSerializer,
                          fields.ListField,
                          relations.ManyRelatedField)):
        return ManyParameter
    elif isinstance(field, serializers.Serializer):
        return SubParameter
    else:
        return Parameter


class Body(object):
    @classmethod
    def from_serializer(cls, serializer, mode='write', **kw):
        params = []
        for name, field in serializer.fields.items():
            param_cls = get_parameter_class(field)
            params.append(param_cls(name, field, mode=mode))
        return cls(params, description=docstring_beautify(type(serializer).__doc__), mode=mode, **kw)

    def __init__(self, parameters, description='', mode='write', partial=False):
        self.parameters = parameters
        self.description = description
        self.mode = mode
        self.partial = partial

    def __repr__(self):
        return 'B{}({!r})'.format(self.mode[0], self.parameters)

    def __iter__(self):
        return iter(self.parameters)

    def __bool__(self):
        return bool(self.parameters)

    def __eq__(self, other):
        return isinstance(other, Body) and self.__dict__ == other.__dict__

    def is_required(self, parameter):
        if self.mode == 'read':
            return False
        if self.partial:
            return False
        return parameter.field.required

    def is_included(self, parameter):
        if self.mode == 'read':
            return not parameter.field.write_only
        else:
            return not parameter.field.read_only

    @property
    def is_complex(self):
        return any(p.is_complex for p in self.parameters if self.is_included(p))

    def get_schema(self):
        return {
            'type': 'object',
            'properties': {
                p.name: p.get_schema() for p in self if self.is_included(p)
            },
            'required': [p.name for p in self if self.is_required(p)],
        }


class Action(object):
    def __init__(self, method, callback, body, responses, overrides=None):
        self.method = method
        self.callback = callback
        self.body = body
        self.responses = responses
        self.overrides = overrides or {}

    def __eq__(self, other):
        return (isinstance(other, Action) and
                self.method == other.method and
                self.callback == other.callback and
                self.body == other.body and
                self.responses == other.responses
                )

    def __repr__(self):
        return '{}<{}>({}; {})'.format(self.method, self.callback, self.body, self.responses)

    @property
    def description(self):
        if 'description' in self.overrides:
            return self.overrides['description']
        return docstring_beautify(self.callback.__doc__)


class PathVariable(object):
    def __init__(self, name, description=None, type='string'):
        self.name = name
        self.description = description or ''
        self.type = type

    def __repr__(self):
        return 'PV({})<{}>'.format(self.name, self.description)

    def __eq__(self, other):
        return isinstance(other, PathVariable) and self.__dict__ == other.__dict__


class Response(object):
    def __init__(self, status, description=None, output=None):
        self.code = status
        self.description = description or responses.get(status)

        if isinstance(output, (serializers.Serializer, serializers.ListSerializer)):
            cls = get_parameter_class(output)
            self.output = cls(None, output, mode='read')
        else:
            self.output = output

    def __eq__(self, other):
        return isinstance(other, Response) and self.__dict__ == other.__dict__

    def __repr__(self):
        return 'R<{}, {} = {}>'.format(self.code, self.description, self.output)

    def get_schema(self):
        if isinstance(self.output, (Body, BaseParameter)):
            return self.output.get_schema()
        return None


class Endpoint(object):
    default_mapping = {
        'get': 'read',
        'post': 'create',
        'put': 'update',
        'patch': 'partial_update',
        'delete': 'destroy',
    }

    def __init__(self, path, callback):
        self.callback = callback
        self.path = path
        self.cls = self.callback.cls

        schema_overrides = getattr(self.cls, 'Schema', None)
        self.overrides = dict(schema_overrides.__dict__) if schema_overrides else {}
        self.overrides['Parameter'] = (dict(self.overrides['Parameter'].__dict__)
                                       if 'Parameter' in self.overrides else {})

    def __repr__(self):
        return '<{}>({})'.format(self.path, self.callback)

    @cached_property
    def path_parameters(self):
        parameters = self.overrides['Parameter']
        pvs = []
        for variable in uritemplate.variables(self.path):
            pvs.append(PathVariable(
                name=variable,
                **parameters.get(variable, {})
            ))
        return pvs

    @property
    def tags(self):
        tags = [x for x in self.path.strip('/').split('/') if '{' not in x]
        if 'tags' in self.overrides:
            tags.extend(self.overrides['tags'])
        return tags

    method_with_body = {'post', 'put', 'patch'}

    def _get_body(self, method):
        if method not in self.method_with_body:
            return Body([])
        serializer_class = self.instance.get_serializer_class()
        return Body.from_serializer(serializer_class(), mode='write', partial=method == 'patch')

    status_code_mapping = {
        'delete': status.HTTP_204_NO_CONTENT,
        'destroy': status.HTTP_204_NO_CONTENT,
        'post': status.HTTP_201_CREATED,
        'create': status.HTTP_201_CREATED,
        'update': status.HTTP_200_OK,
        'partial_update': status.HTTP_200_OK,
        'put': status.HTTP_200_OK,
        'patch': status.HTTP_200_OK,
        'get': status.HTTP_200_OK,
    }

    def _get_responses(self, method):
        overrides = self.overrides.get(method) or {}
        if 'responses' in overrides:
            return [Response(**r) for r in overrides['responses']]

        overrides = overrides.get('response', {})
        status_code = self.status_code_mapping.get(method, status.HTTP_200_OK)

        output = None
        if 'output' in overrides:
            output = overrides['output']
        elif issubclass(self.cls, viewsets.GenericViewSet):
            serializer_class = self.instance.get_serializer_class()
            if method == 'list':
                output = serializers.ListSerializer(child=serializer_class())
            elif method in {'get', 'retrieve', 'update', 'partial_update', 'create', 'put', 'post', 'patch'}:
                output = serializer_class()

        return [
            Response(overrides.get('status', status_code),
                     overrides.get('description'),
                     output)
        ]

    @cached_property
    def instance(self):
        return self.cls()

    @cached_property
    def actions(self):
        """
        Return a descriptive action string for the endpoint, eg. 'list'.
        """
        mapping = getattr(self.callback, 'actions', self.default_mapping)

        if issubclass(self.cls, viewsets.GenericViewSet):
            return [
                Action(http_method,
                       getattr(self.cls, cls_method),
                       self._get_body(http_method),
                       self._get_responses(cls_method),
                       self.overrides.get(cls_method))
                for http_method, cls_method in mapping.items()
                if hasattr(self.cls, cls_method)
            ]
        else:
            return [
                Action(http_method,
                       getattr(self.cls, http_method),
                       self._get_body(http_method),
                       self._get_responses(http_method),
                       self.overrides.get(http_method))
                for http_method in mapping
                if hasattr(self.cls, http_method)
            ]

    @property
    def allowed_methods(self):
        """
        Return a list of the valid HTTP methods for this endpoint.
        """
        if hasattr(self.callback, 'actions'):
            return [
                method.upper() for method in self.callback.actions.keys()
            ]

        return [
            method for method in
            self.callback.cls().allowed_methods if method not in ('OPTIONS', 'HEAD')
        ]

    @property
    def produces(self):
        return {
            r.media_type
            for r in self.instance.get_renderers()
            if r.media_type
        }

    @property
    def consumes(self):
        return {
            r.media_type
            for r in self.instance.get_parsers()
            if r.media_type
        }
