# -*- coding: utf-8 -*-

from importlib import import_module

from django.conf import settings
from django.utils.functional import cached_property
from django.core.urlresolvers import RegexURLPattern, RegexURLResolver
from django.contrib.admindocs.views import simplify_regex

from rest_framework.views import APIView

from autodoc.introspection import Endpoint


def is_api_view(callback):
    """
    Return `True` if the given view callback is a REST framework view/viewset.
    """
    cls = getattr(callback, 'cls', None)
    return (cls is not None) and issubclass(cls, APIView)


class PathExtractor(object):
    def __init__(self, patterns):
        if isinstance(patterns, str):
            patterns = import_module(patterns).urlpatterns
        self.patterns = patterns

    def get_path(self, path_regex):
        """
        Given a URL conf regex, return a URI template string.
        """
        path = simplify_regex(path_regex)
        path = path.replace('<', '{').replace('>', '}')
        return path

    def should_include_endpoint(self, path, callback):
        """
        Return `True` if the given endpoint should be included.
        """
        if not is_api_view(callback):
            return False  # Ignore anything except REST framework views.

        if path.endswith('.{format}') or path.endswith('.{format}/'):
            return False  # Ignore .json style URLs.

        if path == '/':
            return False  # Ignore the root endpoint.
        return True

    def __iter__(self):
        return iter(self._get_api_endpoints(self.patterns))

    def _get_api_endpoints(self, patterns):
        for pattern in patterns:
            path_regex = pattern.regex.pattern
            if isinstance(pattern, RegexURLPattern):
                path = self.get_path(path_regex)
                callback = pattern.callback
                if self.should_include_endpoint(path, callback):
                    yield (path, callback)
            elif isinstance(pattern, RegexURLResolver):
                yield from self._get_api_endpoints(
                    patterns=pattern.url_patterns,
                )


class OpenAPISchemaGenerator(object):
    def __init__(self, info, url, urlconf):
        self.info = info
        self.url = url
        self.urlconf = urlconf

    @cached_property
    def paths(self):
        pe = PathExtractor(self.urlconf)
        return [
            Endpoint(path, callback)
            for path, callback in pe
        ]

    def get_schema(self, request=None):
        all_parameters = [ ]
        if settings.USE_I18N:
            all_parameters.append({
                'name': 'Accept-Language',
                'in': 'header',
                'description': 'The desired language',
                'enum': [code for code, descr in settings.LANGUAGES],
            })
        return {
            'swagger': '2.0',
            'info': self.info,
            'host': request and request.get_host(),
            'basePath': self.url,
            'schemes': [request and request.scheme],
            'consumes': ['application/json'],
            'produces': ['application/json'],
            'paths': self.get_paths(),
            'parameters': all_parameters,
        }

    def get_paths(self):
        return {
            endpoint.path: self.get_path_items(endpoint)
            for endpoint in self.paths
        }

    def get_path_items(self, endpoint):
        return {
            a.method: self.get_operation(a, endpoint)
            for a in endpoint.actions
        }

    def _get_formdata_parameters(self, body):
        for p in body:
            if not body.is_included(p):
                continue
            value = p.get_schema()
            value.update({
                'in': 'formData',
                'required': body.is_required(p),
            })
            yield value

    def _get_body_parameters(self, body):
        return [{
            'name': 'body',
            'in': 'body',
            'description': body.description,
            'required': True,
            'schema': body.get_schema()
        }]

    def get_operation(self, action, endpoint):
        parameters = []
        parameters.extend(self._get_path_fields(endpoint))

        if action.body:
            if action.body.is_complex:
                parameters.extend(self._get_body_parameters(action.body))
            else:
                parameters.extend(self._get_formdata_parameters(action.body))

        responses = self._get_responses(action.responses)

        return {
            'tags': endpoint.tags,
            'description': action.description,
            'consumes': list(sorted(endpoint.consumes)),
            'produces': list(sorted(endpoint.produces)),
            'responses': responses,
            'parameters': list(parameters),
        }

    def _get_responses(self, responses):
        r = {}
        for response in responses:
            values = {
                'description': response.description,
            }
            s = response.get_schema()
            if s:
                values['schema'] = s

            r[response.code] = values
        return r

    def _get_path_fields(self, endpoint):
        for path_param in endpoint.path_parameters:
            yield {
                'name': path_param.name,
                'in': 'path',
                'description': path_param.description,
                'required': True,
                'type': path_param.type,
            }
