# -*- coding: utf-8 -*-

import unittest

from rest_framework import serializers, generics, renderers, parsers, viewsets

from autodoc.introspection import (
    Endpoint,
    PathVariable,
    Action,
    Body,
    Response,
    ManyParameter,
    SubParameter,
)


class MySerializer(serializers.Serializer):
    name = serializers.CharField(
    )
    age = serializers.IntegerField(
        help_text='Number of rotations arround the sun',
    )


class MyApiView(generics.GenericAPIView):
    serializer_class = MySerializer
    renderer_classes = [
        renderers.JSONRenderer,
    ]
    parser_classes = [
        parsers.JSONParser,
    ]

    class Schema:
        class Parameter:
            user_id = {
                'description': 'The id of the user',
            }

        get = {
            'response': {
                'output': MySerializer(),
            }
        }

    def get(self, *args, **kw):
        pass

    def post(self, *args, **kw):
        pass

    def put(self, *args, **kw):
        pass


class MyApiViewset(viewsets.GenericViewSet):
    serializer_class = MySerializer

    class Schema:
        tags = ['foo', 'bar']

    def list(self, *args, **kw):
        pass

    def create(self, *args, **kw):
        pass

    def retrieve(self, *args, **kw):
        pass

    def partial_update(self, *args, **kw):
        pass

    def update(self, *args, **kw):
        pass


class TestBodyApiView(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.ep = Endpoint('schwoob/{user_id}/plym/', MyApiView.as_view())

    def test_api_view(self):
        self.assertEqual(self.ep.path_parameters, [
            PathVariable('user_id', 'The id of the user'),
        ])

    def test_tags(self):
        self.assertListEqual(self.ep.tags, ['schwoob', 'plym'])

    def test_actions(self):
        self.assertListEqual(sorted(self.ep.actions, key=lambda x: x.method), [
            Action('get',
                   MyApiView.get,
                   Body([]),
                   [Response(200, 'OK', SubParameter(None, MySerializer(), mode='read'))]),
            Action('post',
                   MyApiView.post,
                   Body.from_serializer(MySerializer()),
                   [Response(201, 'Created')]),
            Action('put',
                   MyApiView.put,
                   Body.from_serializer(MySerializer()),
                   [Response(200, 'OK')]),
        ])

    def test_allowed_methods(self):
        self.assertSetEqual(set(self.ep.allowed_methods), {'GET', 'POST', 'PUT'})


class TestBodyApiViewsetInstance(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.ep = Endpoint('schwoob/{user_id}/', MyApiViewset.as_view({
            'get': 'retrieve',
            'put': 'update',
            'patch': 'partial_update',
        }))

    def test_tags(self):
        self.assertSetEqual(set(self.ep.tags), {'schwoob', 'foo', 'bar'})

    def test_actions(self):
        self.assertListEqual(sorted(self.ep.actions, key=lambda x: x.method), [
            Action('get',
                   MyApiViewset.retrieve,
                   Body([]),
                   [Response(200, 'OK', SubParameter(None, MySerializer(), mode='read'))]),
            Action('patch',
                   MyApiViewset.partial_update,
                   Body.from_serializer(MySerializer(), partial=True),
                   [Response(200, 'OK', SubParameter(None, MySerializer()))]),
            Action('put',
                   MyApiViewset.update,
                   Body.from_serializer(MySerializer()),
                   [Response(200, 'OK', SubParameter(None, MySerializer()))]),
        ])

    def test_allowed_methods(self):
        self.assertSetEqual(set(self.ep.allowed_methods), {'GET', 'PATCH', 'PUT'})


class TestBodyApiViewsetList(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.ep = Endpoint('schwoob/{user_id}/', MyApiViewset.as_view({
            'get': 'list',
            'post': 'create',
        }))

    def test_actions(self):
        self.assertListEqual(sorted(self.ep.actions, key=lambda x: x.method), [
            Action('get',
                   MyApiViewset.list,
                   Body([]),
                   [Response(200, 'OK', ManyParameter(None, serializers.ListSerializer(child=MySerializer()), mode='read'))]),
            Action('post',
                   MyApiViewset.create,
                   Body.from_serializer(MySerializer()),
                   [Response(201, 'Created', SubParameter(None, MySerializer()))]),
        ])
