# -*- coding: utf-8 -*-


import unittest
from unittest import mock

from rest_framework.routers import SimpleRouter
from rest_framework import viewsets, serializers

from autodoc.schemas import PathExtractor, OpenAPISchemaGenerator


class MySerializer(serializers.Serializer):
    name = serializers.CharField()
    age = serializers.IntegerField()


class MyApiViewset(viewsets.GenericViewSet):
    serializer_class = MySerializer

    def list(self, request, *args, **kw):
        pass

    def retrieve(self, request, *args, **kw):
        pass

    def partial_update(self, request, *args, **kw):
        pass

    def update(self, request, *args, **kw):
        pass


class TestPathExtractor(unittest.TestCase):
    def test_path_extractor_router(self):
        router = SimpleRouter()
        router.register('my', MyApiViewset, base_name='my_thing')

        pe = PathExtractor(router.urls)
        self.assertEqual(list(pe), [
            ('/my/', mock.ANY),
            ('/my/{pk}/', mock.ANY),
        ])


class TestSchemaGenerator(unittest.TestCase):
    def test_schema_generator(self):
        router = SimpleRouter()
        router.register('my', MyApiViewset, base_name='my_thing')

        sg = OpenAPISchemaGenerator({}, '/api/v0/', router.urls)
        schema = sg.get_schema()
        paths = schema.pop('paths')

        assert schema == {
            'swagger': '2.0',
            'basePath': '/api/v0/',
            'host': None,
            'info': {},
            'schemes': [None],
            'parameters': [
                {'enum': ['en', 'fr'], 'in': 'header', 'description': 'The desired language', 'name': 'Accept-Language'}
            ],
            'consumes': ['application/json'],
            'produces': ['application/json'],
        }
        assert paths['/my/{pk}/']['get'] == {
            'responses': {
                200: {
                    'description': 'OK',
                    'schema': {
                        'allowEmptyValue': False,
                        'type': 'object',
                        'description': '',
                        'properties': {
                            'age': {
                                'allowEmptyValue': False,
                                'type': 'integer',
                                'description': '',
                                'name': 'age'},
                            'name': {
                                'allowEmptyValue': False,
                                'type': 'string',
                                'description': '',
                                'name': 'name'}},
                        'required': []
                    }
                }
            },
            'parameters': [{
                'in': 'path',
                'type': 'string',
                'description': '',
                'name': 'pk',
                'required': True
            }],
            'produces': ['application/json', 'text/html'],
            'consumes': ['application/json', 'application/x-www-form-urlencoded', 'multipart/form-data'],
            'tags': ['my'],
            'description': ''
        }
        self.assertDictEqual(paths['/my/']['get'], {
            'responses': {
                200: {
                    'description': 'OK',
                    'schema': {
                        'allowEmptyValue': False,
                        'type': 'array',
                        'items': {
                            'allowEmptyValue': False,
                            'type': 'object',
                            'description': '',
                            'properties': {
                                'age': {
                                    'allowEmptyValue': False,
                                    'type': 'integer',
                                    'description': '',
                                    'name': 'age'
                                },
                                'name': {
                                    'allowEmptyValue': False,
                                    'type': 'string',
                                    'description': '',
                                    'name': 'name'
                                }
                            },
                            'required': []
                        },
                        'description': ''
                    }
                }
            },
            'parameters': [],
            'produces': ['application/json', 'text/html'],
            'consumes': ['application/json', 'application/x-www-form-urlencoded', 'multipart/form-data'],
            'tags': ['my'],
            'description': ''
        })
